/*
 * Lp.h
 * ----
 *
 * the Lp classes
 *
 * Copyright (C) 2004,2007 Christian Keil
 *
 * $Id: Lp.h 506 2007-05-18 13:24:53Z keil $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * */

#ifdef HAVE_GMP
#include <gmp.h>
#endif

#include <map>
#include <vector>
#include <utility>
#include <cstring>
#include <cerrno>
#include <cmath>

enum CoType {
  cotOb,
  cotLe,
  cotGe,
  cotEq
};

#ifdef MALLOC_ALLOC
#define SVECTOR(x) \
  std::vector<std::pair<unsigned int, x>, std::__malloc_alloc_template<0> >
#define ELMMAP(x) \
  std::map<std::pair<unsigned int, unsigned int>, x, ltind, std::__malloc_alloc_template<0> >
#else
#define SVECTOR(x) \
  std::vector<std::pair<unsigned int, x> >
#define ELMMAP(x) \
  std::map<std::pair<unsigned int, unsigned int>, x, ltind>
#endif

/* tests if the first index operand is less than the second */
struct ltind
{
  bool operator()(const std::pair<unsigned int, unsigned int> p1, 
                  const std::pair<unsigned int, unsigned int> p2) const
  {
    if (p1.second < p2.second) return true;
    if ((p1.second == p2.second) && (p1.first < p2.first)) return true;
    return false;
  }
};

struct tConstraint
{
  CoType type;
  unsigned int index;
};

struct tMlSpMatrix
{
  double *pr;
  int *ir, *jc;
  int nz;
};

class CLp {
public:
  CLp() {};
  virtual ~CLp();

  char *name;
  unsigned int m, n, p;
  unsigned int cfree;

  virtual void addObjective(unsigned int i, char *sz) = 0;
  virtual void addIne(unsigned int i, unsigned int j, char *sz) = 0;
  virtual void addIneNeg(unsigned int i, unsigned int j, char *sz) = 0;
  virtual void addEqu(unsigned int i, unsigned int j, char *sz) = 0;

  virtual void allocRhs() = 0;

  virtual void seta(unsigned int i, char *sz) = 0;
  virtual void setaNeg(unsigned int i, char *sz) = 0;
  virtual void setb(unsigned int i, char *sz) = 0;

  virtual void allocBounds() = 0;

  virtual void setxl(unsigned int i, char *sz) = 0;
  virtual void setxu(unsigned int i, char *sz) = 0;
  virtual void setxlFree(unsigned int i) = 0;
  virtual void setxuFree(unsigned int i) = 0;

  virtual void printCoMaps() = 0;
  virtual void printObjective() = 0;
  virtual void printRhs() = 0;
  virtual void printBounds() = 0;
};

#ifdef HAVE_GMP
class CLpExact : public CLp {
public:
  CLpExact() {};
  ~CLpExact();

  tMlSpMatrix A, B;
  mpq_t *a, *b, *c;
  mpq_t *xl, *xu;
  bool *xlInf;
  bool *xuInf;

  SVECTOR(mpq_t *) vObjective;
  ELMMAP(mpq_t *) ineMap;
  ELMMAP(mpq_t *) equMap;

  void addObjective(unsigned int i, char *sz);
  void addIne(unsigned int i, unsigned int j, char *sz);
  void addIneNeg(unsigned int i, unsigned int j, char *sz);
  void addEqu(unsigned int i, unsigned int j, char *sz);

  void allocRhs();

  void seta(unsigned int i, char *sz);
  void setaNeg(unsigned int i, char *sz);
  void setb(unsigned int i, char *sz);

  void allocBounds();

  void setxl(unsigned int i, char *sz);
  void setxu(unsigned int i, char *sz);
  void setxlFree(unsigned int i);
  void setxuFree(unsigned int i);

  void printCoMaps();
  void printObjective();
  void printRhs();
  void printBounds();
};
#endif

class CLpDouble : public CLp {
public:
  CLpDouble() {};
  ~CLpDouble();

  tMlSpMatrix A, B;
  double *a, *b, *c;
  double *xl, *xu;

  SVECTOR(double) vObjective;
  ELMMAP(double) ineMap;
  ELMMAP(double) equMap;

  void addObjective(unsigned int i, char *sz);
  void addIne(unsigned int i, unsigned int j, char *sz);
  void addIneNeg(unsigned int i, unsigned int j, char *sz);
  void addEqu(unsigned int i, unsigned int j, char *sz);

  void allocRhs();

  void seta(unsigned int i, char *sz);
  void setaNeg(unsigned int i, char *sz);
  void setb(unsigned int i, char *sz);

  void allocBounds();

  void setxl(unsigned int i, char *sz);
  void setxu(unsigned int i, char *sz);
  void setxlFree(unsigned int i);
  void setxuFree(unsigned int i);

  void printCoMaps();
  void printObjective();
  void printRhs();
  void printBounds();
};

#ifdef HAVE_GMP
bool parseFloat(char *sz, mpq_t mpq)
{
  /* why can't gmp read floats to rationals? arghh!! */
  int cch = strlen(sz);
  int ex = 0;
  char *szNum = new char[cch + 1];
  memset(szNum, 0, cch + 1);
  char *pch = sz, *pchNum = szNum, *pchRadix = NULL;

#ifdef DEBUG_OUTPUT
  PRINTF("Parsing float string %s\n", sz);
#endif
  while (' ' == *pch) ++pch;
  for (; ('\0' != *pch) && (' ' != *pch); ++pch) {
    if ((('0' <= *pch) && ('9' >= *pch)) || ('+' == *pch) || ('-' == *pch)) {
      *(pchNum++) = *pch;
      continue;
    } else if ('.' == *pch) {
      if (NULL != pchRadix) {
	PRINTF("parseFloat: WARNING: Found second radix point in %s\n", sz);
	return false;
      }
      pchRadix = pchNum;
      continue;
    } else if (('e' == *pch) || ('E' == *pch)) {
      errno = 0;
      ex = strtol(pch + 1, NULL, 10);
      if (0 != errno) {
	PRINTF("parseFloat: ERROR: Converting exponent to int int %s\n", sz);
	return false;
      }
      break;
    } 
    PRINTF("parseFloat: ERROR: Unexpected character '%c' in float %s\n", *pch, sz);
  }

#ifdef DEBUG_OUTPUT
  PRINTF("Parsed to numerator string %s\n", szNum);
  if (NULL != pchRadix) {
    PRINTF("Radix point present at index %d\n", pchRadix - szNum);
  }
  PRINTF("Parsed to exponent %ld\n", ex);
#endif

  if ((NULL != pchRadix) && (pchRadix < pchNum)) {
    ex += pchRadix - pchNum;
  }
  mpz_set_str(mpq_numref(mpq), szNum, 10);
  mpz_set_si(mpq_denref(mpq), 1);
  if (ex > 0) {
    for (int i = 0; i < ex; ++i) mpz_mul_ui(mpq_numref(mpq), mpq_numref(mpq), 10);
  } else if (ex < 0) {
    for (int i = 0; i > ex; --i) mpz_mul_ui(mpq_denref(mpq), mpq_denref(mpq), 10);
  }
  mpq_canonicalize(mpq);
  delete [] szNum;

#ifdef DEBUG_OUTPUT
  PRINTF("Transformed to rational ");
  mpq_out_str(NULL, 10, mpq);
  PRINTF("\n");
#endif
  
  return true;
}
#endif

CLp::~CLp() 
{
  FREE(name);
}

#ifdef HAVE_GMP
CLpExact::~CLpExact()
{
  for (ELMMAP(mpq_t *)::iterator it = ineMap.begin();
       it != ineMap.end();
       ++it) {
    mpq_clear(*(it->second));
  }
  for (ELMMAP(mpq_t *)::iterator it = equMap.begin();
       it != equMap.end();
       ++it) {
    mpq_clear(*(it->second));
  }
  for (SVECTOR(mpq_t *)::iterator it = vObjective.begin();
       it != vObjective.end();
       ++it) {
    mpq_clear(*(it->second));
  }
  ineMap.clear();
  equMap.clear();
  vObjective.clear();

  /* We don't use CLpExact in Matlab so we can free these normally */
  for (int i = 0; i < m; ++i) {
    mpq_clear(a[i]);
  }
  free(a);
  for (int i = 0; i < p; ++i) {
    mpq_clear(b[i]);
  }
  free(b);
  for (int i = 0; i < n; ++i) {
    mpq_clear(xl[i]);
    mpq_clear(xu[i]);
  }
  free(xl);
  free(xu);
}

void
CLpExact::addObjective(unsigned int i, char *sz)
{
  mpq_t *pmpq = reinterpret_cast<mpq_t *>(new mpq_t);
  mpq_init(*pmpq);
  parseFloat(sz, *pmpq);

  vObjective.push_back(std::make_pair(i, pmpq));
}

void
CLpExact::addIne(unsigned int i, unsigned int j, char *sz)
{
  mpq_t *pmpq = reinterpret_cast<mpq_t *>(new mpq_t);
  mpq_init(*pmpq);
  parseFloat(sz, *pmpq);
  ineMap[std::make_pair(i, j)] = pmpq;
}

void
CLpExact::addIneNeg(unsigned int i, unsigned int j, char *sz)
{
  mpq_t *pmpq = reinterpret_cast<mpq_t *>(new mpq_t);
  mpq_init(*pmpq);
  parseFloat(sz, *pmpq);
  mpq_neg(*pmpq, *pmpq);
  ineMap[std::make_pair(i, j)] = pmpq;
}

void
CLpExact::addEqu(unsigned int i, unsigned int j, char *sz)
{
  mpq_t *pmpq = reinterpret_cast<mpq_t *>(new mpq_t);
  mpq_init(*pmpq);
  parseFloat(sz, *pmpq);
  equMap[std::make_pair(i, j)] = pmpq;
}

void
CLpExact::allocRhs()
{
  /* We don't use CLpExact in Matlab so we can malloc these normally */
  a = (mpq_t *) malloc(m * sizeof (mpq_t));
  for (int i = 0; i < m; ++i) {
    mpq_init(a[i]);
  }
  b = (mpq_t *) malloc(p * sizeof (mpq_t));
  for (int i = 0; i < p; ++i) {
    mpq_init(b[i]);
  }
}

void
CLpExact::seta(unsigned int i, char *sz)
{
  parseFloat(sz, a[i-1]);
}

void
CLpExact::setaNeg(unsigned int i, char *sz)
{
  parseFloat(sz, a[i-1]);
  mpq_neg(a[i-1], a[i-1]);
}

void
CLpExact::setb(unsigned int i, char *sz)
{
  parseFloat(sz, b[i-1]);
}

void
CLpExact::allocBounds()
{
  /* We don't use CLpExact in Matlab so we can malloc these normally */
  xl = (mpq_t *) malloc(n * sizeof (mpq_t));
  xu = (mpq_t *) malloc(n * sizeof (mpq_t));
  xlInf = (bool *) malloc(n * sizeof (bool));
  xuInf = (bool *) malloc(n * sizeof (bool));
  for (int i = 0; i < n; ++i) {
    mpq_init(xl[i]);
    xlInf[i] = false;
    mpq_init(xu[i]);
    xuInf[i] = true;
  }
}

void
CLpExact::setxl(unsigned int i, char *sz)
{
  parseFloat(sz, xl[i-1]);
}

void
CLpExact::setxu(unsigned int i, char *sz)
{
  parseFloat(sz, xu[i-1]);
}

void
CLpExact::setxlFree(unsigned int i)
{
  xlInf[i-1] = true;
}

void
CLpExact::setxuFree(unsigned int i)
{
  xuInf[i-1] = true;
}

void
CLpExact::printCoMaps()
{
  PRINTF("Inequalities: %d matrix entries\n", ineMap.size());
  for (ELMMAP(mpq_t *)::iterator it = ineMap.begin();
       it != ineMap.end();
       ++it) {
    PRINTF("[%d, %d] => {", it->first.first, it->first.second);
    mpq_out_str(NULL, 10, *(it->second));
    PRINTF("}\n");
  }

  PRINTF("Equations: %d matrix entries\n", equMap.size());
  for (ELMMAP(mpq_t *)::iterator it = equMap.begin();
       it != equMap.end();
       ++it) {
    PRINTF("[%d, %d] => {", it->first.first, it->first.second);
    mpq_out_str(NULL, 10, *(it->second));
    PRINTF("}\n");
  }
}

void
CLpExact::printObjective()
{
  PRINTF("Objective: %d vector elements\n", vObjective.size());
  for (SVECTOR(mpq_t *)::iterator it = vObjective.begin();
      it != vObjective.end();
      ++it) {
    PRINTF("{%d, ", it->first);
    mpq_out_str(NULL, 10, *(it->second));
    PRINTF("}\n");
  }
}

void
CLpExact::printRhs()
{
  PRINTF("RHS:\n");
  PRINTF("a\n");
  for (int i = 0; i < m; ++i) {
    PRINTF("[%d] => ", i);
    mpq_out_str(NULL, 10, a[i]);
    PRINTF("\n");
  }
  PRINTF("b\n");
  for (int i = 0; i < p; ++i) {
    PRINTF("[%d] => ", i);
    mpq_out_str(NULL, 10, b[i]);
    PRINTF("\n");
  }
}

void
CLpExact::printBounds()
{
  PRINTF("Simple bounds:\n");
  for (int i = 0; i < n; ++i) {
    PRINTF("[%d] => [", i);
    if (xlInf[i]) PRINTF("-oo, ");
    else {
      mpq_out_str(NULL, 10, xl[i]);
      PRINTF(", ");
    }
    if (xuInf[i]) PRINTF("oo]\n");
    else {
      mpq_out_str(NULL, 10, xu[i]);
      PRINTF("]\n");
    }
  }
}
#endif

CLpDouble::~CLpDouble()
{
  ineMap.clear();
  equMap.clear();
  vObjective.clear();

#ifndef MEX
  /* Matlab takes care of this itself */
  free(xl);
  free(xu);
#endif
}

void
CLpDouble::addObjective(unsigned int i, char *sz)
{
  vObjective.push_back(std::make_pair(i, atof(sz)));
}

void
CLpDouble::addIne(unsigned int i, unsigned int j, char *sz)
{
  ineMap[std::make_pair(i, j)] = atof(sz);
}

void
CLpDouble::addIneNeg(unsigned int i, unsigned int j, char *sz)
{
  ineMap[std::make_pair(i, j)] = -atof(sz);
}

void
CLpDouble::addEqu(unsigned int i, unsigned int j, char *sz)
{
  equMap[std::make_pair(i, j)] = atof(sz);
}

void
CLpDouble::allocRhs()
{
  a = (double *) MALLOC(m * sizeof (double));
  memset(a, 0, m * sizeof (double));
  b = (double *) MALLOC(p * sizeof (double));
  memset(b, 0, p * sizeof (double));
}

void
CLpDouble::seta(unsigned int i, char *sz)
{
  a[i-1] = atof(sz);
}

void
CLpDouble::setaNeg(unsigned int i, char *sz)
{
  a[i-1] = -atof(sz);
}

void
CLpDouble::setb(unsigned int i, char *sz)
{
  b[i-1] = atof(sz);
}

void
CLpDouble::allocBounds()
{
  xl = (double *) MALLOC(n * sizeof (double));
  memset(xl, 0, n * sizeof (double));
  xu = (double *) MALLOC(n * sizeof (double));
  for (int i = 0; i < n; ++i) {
    xu[i] = (double) INFINITY;
  }
}

void
CLpDouble::setxl(unsigned int i, char *sz)
{
  xl[i-1] = atof(sz);
}

void
CLpDouble::setxu(unsigned int i, char *sz)
{
  xu[i-1] = atof(sz);
}

void
CLpDouble::setxlFree(unsigned int i)
{
  xl[i-1] = -INFINITY;
}

void
CLpDouble::setxuFree(unsigned int i)
{
  xu[i-1] = INFINITY;
}

void
CLpDouble::printCoMaps()
{
  PRINTF("Inequalities: %d matrix entries\n", ineMap.size());
  for (ELMMAP(double)::iterator it = ineMap.begin();
       it != ineMap.end();
       ++it)
    PRINTF("[%d, %d] => {%e}\n", it->first.first, it->first.second, it->second);

  PRINTF("Equations: %d matrix entries\n", equMap.size());
  for (ELMMAP(double)::iterator it = equMap.begin();
       it != equMap.end();
       ++it)
    PRINTF("[%d, %d] => {%e}\n", it->first.first, it->first.second, it->second);
}

void
CLpDouble::printObjective()
{
  PRINTF("Objective: %d vector elements\n", vObjective.size());
  for (SVECTOR(double)::iterator it = vObjective.begin();
      it != vObjective.end();
      ++it) {
    PRINTF("{%d, %e}\n", it->first, it->second);
  }
}

void
CLpDouble::printRhs()
{
  PRINTF("RHS:\n");
  PRINTF("a\n");
  for (int i = 0; i < m; ++i) {
    PRINTF("[%d] => %e\n", i, a[i]);
  }
  PRINTF("b\n");
  for (int i = 0; i < p; ++i) {
    PRINTF("[%d] => %e\n", i, b[i]);
  }
}

void
CLpDouble::printBounds()
{
  PRINTF("Simple bounds:\n");
  for (int i = 0; i < n; ++i) {
    PRINTF("[%d] => [%e, %e]\n", i, xl[i], xu[i]);
  }
}
