#include <mex.h>
#include <matrix.h>
#include <mat.h>

#define MEX
#include "readMps.h"

void
mexFunction(int nlhs, mxArray *plhs[],
            int nrhs, const mxArray *prhs[])
{
  bool fSilent = false;

  if (nrhs == 2) {
    if (!(mxIsLogical(prhs[1]) || (mxGetNumberOfElements(prhs[1]) != 1))) {
      mexErrMsgTxt("\tSecond input argument (SILENT) must be logical scalar\n");
    }

    fSilent = mxGetLogicals(prhs[1])[0];
  }

  if (!fSilent) {
    mexPrintf("\to========================================o\n"
	      "\t|         MPS Parser for MATLAB          |\n"
	      "\t|       ,-----------------------'        |\n"
	      "\t|                                        |\n"
	      "\t| %38s |\n"
	      "\t| %38s |\n"
	      "\t`========================================'\n\n",
	      copyright, revision);
  }

  char *usage = "\tusage:\n"
		"\t  mps2mat(<path/to/mpsfile>)\n\n"
		"\tsee help mps2mat\n";

  /* check number of right hand side arguments */
  if ((nrhs < 1) || (nrhs > 2)) {
    mexPrintf("%s", usage);
    mexErrMsgTxt("Wrong number of input arguments");
  }

  /* check if first right hand side argument can be a path */
  if (!mxIsChar(prhs[0])) {
    mexPrintf("%s", usage);
    mexErrMsgTxt("First input argument (<path/to/mpsfile>) must be a string");
  }
  
  /* check number of left hand arguments */
  if (nlhs != 0) {
    mexErrMsgTxt("No output arguments allowed\n");
  }

  /* parse mps file */
  CLpDouble lp;
  unsigned int buflen = mxGetM(prhs[0]) * mxGetN(prhs[0]) * sizeof(mxChar)+ 1;
  char *path = (char *) MALLOC(buflen);
  char *matPath = NULL;
  mxGetString(prhs[0], path, buflen);

  /* construct path to MAT file */
  matPath = (char *) MALLOC(buflen);
  memset(matPath, 0, buflen);
  strncpy(matPath, path, strlen(path) - 3);
  strcat(matPath, "mat");

  /* test if file already exists */
  MATFile *out = matOpen(matPath, "r");
  if (out != NULL) {
    mxArray *inParm[2];
    mxArray *outParm[1];

    char *query = (char *) MALLOC(strlen(matPath) + 36 + 1);
    sprintf(query, "%s already exists - overwrite? [y/n]: ", matPath);
    inParm[0] = mxCreateString(query);
    inParm[1] = mxCreateString("s");

    mexCallMATLAB(1, outParm, 2, inParm, "input");

    if ((mxGetChars(outParm[0])[0] != 'y') && 
	(mxGetChars(outParm[0])[0] != 'Y')) {
      matClose(out);
      return;
    }
    matClose(out);
  }

  if (readMps(path, lp, fSilent) == -1) {
    mexErrMsgTxt("Could not parse MPS file");
  }
  mxFree(path);

  buildABc(lp);

  out = matOpen(matPath, "w");
  if (!out) {
    mexErrMsgTxt("Could not open MAT file for writing\n");
  }

  mxArray *write = NULL;
  write = mxCreateSparse(lp.m, lp.n, lp.A.nz, mxREAL);
  mxSetPr(write, lp.A.pr);
  mxSetIr(write, lp.A.ir);
  mxSetJc(write, lp.A.jc);
  if (matPutVariable(out, "A", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write A to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateSparse(lp.p, lp.n, lp.B.nz, mxREAL);
  mxSetPr(write, lp.B.pr);
  mxSetIr(write, lp.B.ir);
  mxSetJc(write, lp.B.jc);
  if (matPutVariable(out, "B", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write B to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateDoubleMatrix(lp.m, 1, mxREAL);
  mxSetPr(write, lp.a);
  if (matPutVariable(out, "a", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write a to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateDoubleMatrix(lp.p, 1, mxREAL);
  mxSetPr(write, lp.b);
  if (matPutVariable(out, "b", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write b to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateDoubleMatrix(lp.n, 1, mxREAL);
  mxSetPr(write, lp.c);
  if (matPutVariable(out, "c", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write c to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateDoubleMatrix(lp.n, 1, mxREAL);
  mxSetPr(write, lp.xl);
  if (matPutVariable(out, "xl", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write xl to MAT file");
  }
  mxDestroyArray(write);

  write = mxCreateDoubleMatrix(lp.n, 1, mxREAL);
  mxSetPr(write, lp.xu);
  if (matPutVariable(out, "xu", write)) {
    if (ferror(matGetFp(out)))
      mexPrintf("%s\n", strerror(errno));
    matClose(out);
    mexErrMsgTxt("Could not write xu to MAT file");
  }
  mxDestroyArray(write);

  if (matClose(out)) {
    if (ferror(matGetFp(out))) {
      mexPrintf("%s\n", strerror(errno));
    }
    mexErrMsgTxt("Could not close MAT file");
  }
}
