/*
 * readMps.h
 * ---------
 *
 * Matlab Mex routine to parse mps files.
 *
 * Copyright (C) 2006,2007 Christian Keil
 *
 * $Id: readMps.h 509 2007-05-24 14:52:14Z keil $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * */
#ifndef READMPS_H
#define READMPS_H

static char *revision = "$Revision: 509 $";
static const char *copyright = "Copyright (C) 2006,2007 Christian Keil";


/* generate debug output (mainly print the generated intermediate structures -
 * maps / vectors of coefficients) */
#undef DEBUG_OUTPUT
/* use malloc to allocate objects in the maps / vectors usefull for leak
 * checking - may not be portable to future gcc versions!!! */
#undef MALLOC_ALLOC

#ifdef MEX
#define PRINTF mexPrintf
#define CALLOC mxCalloc
#define MALLOC mxMalloc
#define FREE   mxFree
#else
#define PRINTF printf
#define CALLOC calloc
#define MALLOC malloc
#define FREE   free
#endif

#include "Lp.h"
#include <map>
#include <vector>
#include <utility>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <errno.h>

/* tests if the first string operand is less than the second */
struct ltstr
{
  bool operator()(const char *s1, const char *s2) const
  {
    return std::strcmp(s1, s2) < 0;
  }
};

#ifdef MALLOC_ALLOC
typedef std::vector<char, std::__malloc_alloc_template<0> > tCharVec;

typedef std::map<const char *, tConstraint, ltstr,
  std::__malloc_alloc_template<0> > tConMap;

typedef std::map<const char *, unsigned int, ltstr,
  std::__malloc_alloc_template<0> > tVarMap;
#else
typedef std::vector<char> tCharVec;

typedef std::map<const char *, tConstraint, ltstr> tConMap;

typedef std::map<const char *, unsigned int, ltstr> tVarMap;
#endif

#ifdef DEBUG_OUTPUT
void
printConMap(tConMap conMap)
{
  PRINTF("%d constraints\n", conMap.size());
  PRINTF("[name] => {type, index}\n");
  for (tConMap::iterator it = conMap.begin();
       it != conMap.end();
       ++it)
    PRINTF("[%s] => {%d, %d}\n", it->first, it->second.type,
	it->second.index);
}

void
printVarMap(tVarMap varMap)
{
  PRINTF("%d variables\n", varMap.size());
  PRINTF("[name] => {index}\n");
  for (tVarMap::iterator it = varMap.begin();
       it != varMap.end();
       ++it) {
    PRINTF("[%s] => {%d}\n", it->first, it->second);
  }
}

void
printArray(double *a, unsigned int dim)
{
  PRINTF("%d array elements\n", dim);
  for (int i = 0; i < dim; ++i)
    PRINTF("[%d] => %e\n", i, a[i]);
}
#endif

char *
getLine(char *sz, FILE *in, bool fSilent)
{
  /* We print informative messages once if we encounter the input file to
   * possibly be in free form or we encounter the input file to contain
   * comments. */
  static bool fFreeForm = false;
  static bool fCommentFound = false;
  static int csz = 0;

  if (in == NULL) {
    fFreeForm = false;
    fCommentFound = false;
    csz = 0;
    return NULL;
  } 
    
  do {
    if (fgets(sz, 102, in) == NULL) {
      if (feof(in)) {
	PRINTF("ERROR: unexpected EOF encountered\n");
	sz = NULL;
      } else {
	PRINTF("ERROR: %s\n", strerror(errno));
	sz = NULL;
      }
      return NULL;
    }
    ++csz;

    char *pch = strchr(sz, '\n');
    if (pch == NULL) {
      if (!fSilent) PRINTF("truncating line %d after 100 characters\n", csz);
      sz[100] = '\n';
      pch = sz + 100;
      while (fgetc(in) != '\n') {}
    }
    
    if (sz[0] == '*') {
      if (!fCommentFound) {
	fCommentFound = true;
	if (!fSilent) PRINTF("lines starting with a '*' are comments from the mps file\n");
      }
      if (!fSilent) PRINTF("%s", sz);
    } else if ((!fFreeForm) && (pch - sz > 61)) {
      fFreeForm = true;
      PRINTF("\nWarning: MPS seems to be free form which is not handled by "
		"readMps !!!\n\n");
    }
  } while (sz[0] == '*');

  if (sz[strlen(sz) - 1] == '\n') sz[strlen(sz) - 1] = '\0';

#ifdef DEBUG_OUTPUT
  PRINTF("read line: %s\n", sz);
#endif

  return sz;
}

bool
checkBoundRow(char *line, char tokensRead, char tokensRequired,
              char *bndVar, tVarMap &varMap, tVarMap::iterator &varIt)
{
  if (tokensRead != tokensRequired) {
    PRINTF("ERROR! wrong number of tokens in bound row in line:\n"
	      "%s\n", line);
    return false;
  }
  if ((varIt = varMap.find(strtok(bndVar, " "))) == varMap.end()) {
    PRINTF("ERROR! unknown variable %s in line:\n"
	      "%s\n", strtok(bndVar, " "), line);
    return false;
  }
  return true;
}
  
int 
readMps(FILE *in, CLp &lp, bool fSilent = false)
{
  /*
   * steps to perform:
   * 1) NAME section
   *    a) read name of problem
   * 2) ROWS section
   *    a) read objective function/constraint names and build structure
   *       containing names and types 
   * 3) COLUMNS section
   *    a) read column names and build structure containing names
   *    b) match row names to structure from 2a) and store values in appropriate
   *       matrix
   * 4) RHS section
   *    a) match constraint name to structure from 2a) and store right hand side
   *       in array
   * 5) BOUNDS section
   *    a) match column names to structure from 3a) and store bounds in
   *       array
   * [...]
   * 6) read ENDATA
   * */

  /* We have to read a whole line with 61 characters plus the terminating '\n'
   * from the input file and terminate the resulting string with an additional
   * '\0'. As the mps files seem to have quite longer lines lets extend this to
   * 100 characters. */
  char line[100 + 2];
  /* Reset getLine status */
  getLine(NULL, NULL, fSilent);
  /* Loop until we find a line which doesn't start with a * - these are comment
   * lines */
  if (!getLine(line, in, fSilent)) return -1;

  /* 1) NAME section
   *    a) read name of problem
   * */
  if (NULL == (lp.name = (char *) MALLOC(8 + 1))) {
    PRINTF("ERROR: Out of memory\n");
    return -1;
  }
  memset(lp.name, 0, 9);
  if (strncmp("NAME ", line, 5) != 0) {
    PRINTF("ERROR: NAME section expected in line:\n"
	   "%s\n", line);
  }
  if (sscanf(line, "NAME%*10c%8c", lp.name) != 1) {
    /* <lp.name> not successfully read - line malformed */
    PRINTF("WARNING: NAME line malformed:\n"
	   "%s\n", line);
    PRINTF("MPS may be free form, which is not handled by readMps.\n");
    strncpy(lp.name, line + 14, 8);
  }
  if (!fSilent) PRINTF("NAME\tread: %s\n", lp.name);

  if (!getLine(line, in, fSilent)) return -1;

  /* 2) ROWS section
   *    a) read objective function/constraint names and build structure
   *       containing names and types 
   * */
  if (std::strncmp(line, "ROWS", 4) != 0) {
    /* ROWS section should start with "ROWS" in column 1 */
    PRINTF("ERROR! ROWS section expected in line:\n"
	      "%s\n", line);
    return -1;
  }
  lp.m = 0;
  lp.p = 0;
  bool objFound = false;
  /* Compiler complains if the key type is an array as these can't be assigned
   * to in ISO C++. Let's make it a <const char *> and make sure the keys aren't
   * longer than 8 chars by using the field width in scanning. 
   * This should be some more efficient structure to find elements using their
   * key like a hash. This is not included in the supplied header files so
   * perhaps we'll switch to the "original" stl later... */
  tConMap conMap;
  if (!getLine(line, in, fSilent)) return -1;
  char conType[2 + 1] = {};
  char conName[8 + 1];
  tConstraint newCon;
  while (line[0] == ' ') {
    /* only recognizes constraint types with the letter in column 2! mps format
     * allows it also in column 3... */
    memset(conName, 0, 9);
    if (sscanf(line, "%*1c%2c%*1c%8c", conType, conName) != 2) {
      PRINTF("ERROR! wrong number of tokens in constraint line:\n"
	        "%s\n", line);
      return -1;
    }
    if ((0 == strcmp(conType, "N ")) || (0 == strcmp(conType, " N"))) {
      if (!objFound) {
	objFound = true;
	newCon.type = cotOb;
	newCon.index = 0;
	conMap[strcpy(new char[8 + 1], strtok(conName, " "))] = newCon;
      } else {
	/* objective function already found */
	PRINTF("Ignoring additional non constraing row in line:\n"
	       "%s\n", line);
      }
    } else if ((0 == strcmp(conType, "L ")) || (0 == strcmp(conType, " L"))) {
      newCon.type = cotLe;
      newCon.index = ++lp.m;
      conMap[strcpy(new char[8 + 1], strtok(conName, " "))] = newCon;
    } else if ((0 == strcmp(conType, "G ")) || (0 == strcmp(conType, " G"))) {
      newCon.type = cotGe;
      newCon.index = ++lp.m;
      conMap[strcpy(new char[8 + 1], strtok(conName, " "))] = newCon;
    } else if ((0 == strcmp(conType, "E ")) || (0 == strcmp(conType, " E"))) {
      newCon.type = cotEq;
      newCon.index = ++lp.p;
      conMap[strcpy(new char[8 + 1], strtok(conName, " "))] = newCon;
    } else {
      PRINTF("WARNING: ignoring unknown constraint type %c in line:\n"
	     "%s\n", conType, line);
    }
    if (!getLine(line, in, fSilent)) return -1;
  }
  if (!objFound) {
    PRINTF("ERROR: no objective function found");
    return -1;
  }
  if (!fSilent) PRINTF("ROWS\tread: %d equalit%s, %d inequalit%s\n", 
		       lp.p, lp.p == 1 ? "y" : "ies",
		       lp.m, lp.m == 1 ? "y" : "ies");
#ifdef DEBUG_OUTPUT
  printConMap(conMap);
#endif
    
  /* 3) COLUMNS section
   *    a) read column names and build structure containing names
   *    b) match row names to structure from 2a) and store values in appropriate
   *       matrix
   * */
  if (std::strncmp(line, "COLUMNS", 7) != 0) {
    /* COLUMNS section should start with "COLUMNS" in column 1 */
    PRINTF("ERROR! COLUMNS section expected in line:\n"
	      "%s\n", line);
    return -1;
  }
  lp.n = 0;
  unsigned int thisVar = 0;
  tVarMap varMap;
  tVarMap::iterator varIt;
  char varName[8 + 1];
  char varConName1[8 + 1];
  char varConName2[8 + 1];
  char varConVal1[12 + 1];
  char varConVal2[12 + 1];
  tConMap::iterator conIt;
  bool moreCols = true;
  while (moreCols) {
    if (!getLine(line, in, fSilent)) return -1;
    memset(varName, 0, 9);
    memset(varConName1, 0, 9);
    memset(varConVal1, 0, 13);
    memset(varConName2, 0, 9);
    memset(varConVal2, 0, 13);
    switch (sscanf(line, "%*4c%8c%*2c%8c%*2c%12c%*3c%8c%*2c%12c",
	           varName, varConName1, varConVal1, varConName2, varConVal2)) {
      case 3:
	if ((varIt = varMap.find(strtok(varName, " "))) == varMap.end()) {
	  thisVar = ++lp.n;
	  varMap[strcpy(new char[8 + 1], strtok(varName, " "))] = thisVar;
	} else {
	  thisVar = varIt->second;
	}
	if ((conIt = conMap.find(strtok(varConName1, " "))) == conMap.end()) {
	  PRINTF("ERROR: unknown row %s in line:\n"
	            "%s\n", strtok(varConName1, " "), line);
	  return -1;
	}
	switch (conIt->second.type) {
	  case cotOb:
	    lp.addObjective(thisVar, varConVal1);
	    break;
	  case cotLe:
	    lp.addIne(conIt->second.index, thisVar, varConVal1);
	    break;
	  case cotGe:
	    lp.addIneNeg(conIt->second.index, thisVar, varConVal1);
	    break;
	  case cotEq:
	    lp.addEqu(conIt->second.index, thisVar, varConVal1);
	    break;
	}
	break;
      case 5:
	if ((varIt = varMap.find(strtok(varName, " "))) == varMap.end()) {
	  thisVar = ++lp.n;
	  varMap[strcpy(new char[8 + 1], strtok(varName, " "))] = thisVar;
	} else
	  thisVar = varIt->second;
	if ((conIt = conMap.find(strtok(varConName1, " "))) == conMap.end()) {
	  PRINTF("ERROR! unknown row %s in line:\n"
	            "%s\n", strtok(varConName1, " "), line);
	  return -1;
	}
	switch (conIt->second.type) {
	  case cotOb:
	    lp.addObjective(thisVar, varConVal1);
	    break;
	  case cotLe:
	    lp.addIne(conIt->second.index, thisVar, varConVal1);
	    break;
	  case cotGe:
	    lp.addIneNeg(conIt->second.index, thisVar, varConVal1);
	    break;
	  case cotEq:
	    lp.addEqu(conIt->second.index, thisVar, varConVal1);
	    break;
	}
	if ((conIt = conMap.find(strtok(varConName2, " "))) == conMap.end()) {
	  PRINTF("ERROR! unknown row %s in line:\n"
	            "%s\n", strtok(varConName2, " "), line);
	  return -1;
	}
	switch (conIt->second.type) {
	  case cotOb:
	    lp.addObjective(thisVar, varConVal2);
	    break;
	  case cotLe:
	    lp.addIne(conIt->second.index, thisVar, varConVal2);
	    break;
	  case cotGe:
	    lp.addIneNeg(conIt->second.index, thisVar, varConVal2);
	    break;
	  case cotEq:
	    lp.addEqu(conIt->second.index, thisVar, varConVal2);
	    break;
	}
	break;
      default:
	moreCols = false;
    }
  }
  if (!fSilent) PRINTF("COLUMNS\tread: %d variables\n", lp.n);
#ifdef DEBUG_OUTPUT
  printVarMap(varMap);
  lp.printObjective();
  lp.printCoMaps();
#endif

  /* 4) RHS section
   *    a) match constraint name to structure from 2a) and store right hand side
   *       in array
   * */
  lp.allocRhs();
  if (std::strncmp(line, "RHS", 3) != 0) {
    if (!fSilent) PRINTF("no RHS section in file\n");
  } else {
    int tokens;
    char rhsVarName1[8 + 1];
    char rhsVarName2[8 + 1];
    char rhsVal1[12 + 1];
    char rhsVal2[12 + 1];
    bool moreRhs = true;
    while (moreRhs) {
      if (!getLine(line, in, fSilent)) return -1;
      memset(rhsVarName1, 0, 9);
      memset(rhsVal1, 0, 13);
      memset(rhsVarName2, 0, 9);
      memset(rhsVal2, 0, 13);
      switch (sscanf(line, "%*4c%*8c%*2c%8c%*2c%12c%*3c%8c%*2c%12c",
	             rhsVarName1, rhsVal1, rhsVarName2, rhsVal2)) {
	case 2:
	  if ((conIt = conMap.find(strtok(rhsVarName1, " "))) == conMap.end()) {
	    PRINTF("ERROR! unknown row %s in line:\n"
		"%s\n", strtok(rhsVarName1, " "), line);
	    return -1;
	  }
	  switch (conIt->second.type) {
	    case cotOb:
	      PRINTF("WARNING! Ignoring right hand side value for"
		        " objective function\n");
	      break;
	    case cotLe:
	      lp.seta(conIt->second.index, rhsVal1);
	      break;
	    case cotGe:
	      lp.setaNeg(conIt->second.index, rhsVal1);
	      break;
	    case cotEq:
	      lp.setb(conIt->second.index, rhsVal1);
	      break;
	  }
	  break;
	case 4:
	  if ((conIt = conMap.find(strtok(rhsVarName1, " "))) == conMap.end()) {
	    PRINTF("ERROR! unknown row %s in line:\n"
		"%s\n", strtok(rhsVarName1, " "), line);
	    return -1;
	  }
	  switch (conIt->second.type) {
	    case cotOb:
	      PRINTF("WARNING! Ignoring right hand side value for"
		        " objective function\n");
	      break;
	    case cotLe:
	      lp.seta(conIt->second.index, rhsVal1);
	      break;
	    case cotGe:
	      lp.setaNeg(conIt->second.index, rhsVal1);
	      break;
	    case cotEq:
	      lp.setb(conIt->second.index, rhsVal1);
	      break;
	  }
	  if ((conIt = conMap.find(strtok(rhsVarName2, " "))) == conMap.end()) {
	    PRINTF("ERROR! unknown row %s in line:\n"
		"%s\n", strtok(rhsVarName2, " "), line);
	    return -1;
	  }
	  switch (conIt->second.type) {
	    case cotOb:
	      PRINTF("WARNING! Ignoring right hand side value for"
		        " objective function\n");
	      break;
	    case cotLe:
	      lp.seta(conIt->second.index, rhsVal2);
	      break;
	    case cotGe:
	      lp.setaNeg(conIt->second.index, rhsVal2);
	      break;
	    case cotEq:
	      lp.setb(conIt->second.index, rhsVal2);
	      break;
	  }
	  break;
	default:
	  moreRhs = false;
      }
    }
    if (!fSilent) PRINTF("RHS\tread\n");
  }
#ifdef DEBUG_OUTPUT
  lp.printRhs();
#endif

  /* 5) BOUNDS section
   *    a) match column names to structure from 3a) and store bounds in
   *       appropriate vector elements
   * */
  lp.allocBounds();
  lp.cfree = 0;
  if (std::strncmp(line, "BOUNDS", 6) != 0) {
    if (!fSilent) PRINTF("no BOUNDS section in file\n");
  } else {
    unsigned int fixed = 0;
    if (!getLine(line, in, fSilent)) return -1;
    char bndType[2 + 1];
    char bndVar[8 + 1];
    char bndVal[12 + 1];
    int tokens = 0;
    while (line[0] == ' ') {
      memset(bndType, 0, 3);
      memset(bndVar, 0, 9);
      memset(bndVal, 0, 13);
      tokens = sscanf(line, "%*1c%2c%*1c%*8c%*2c%8c%*2c%12c", 
	              bndType, bndVar, bndVal);
      if (strcmp(bndType, "LO") == 0) {
	/* lower bound */
	if (!checkBoundRow(line, tokens, 3, bndVar, varMap, varIt))
	  return -1;
	else
	  lp.setxl(varIt->second, bndVal);
      } else if (strcmp(bndType, "UP") == 0) {
	/* upper bound */
	if (!checkBoundRow(line, tokens, 3, bndVar, varMap, varIt))
	  return -1;
	else
	  lp.setxu(varIt->second, bndVal);
      } else if (strcmp(bndType, "FX") == 0) {
	/* fixed variable */
	if (!checkBoundRow(line, tokens, 3, bndVar, varMap, varIt))
	  return -1;
	else {
	  lp.setxl(varIt->second, bndVal);
	  lp.setxu(varIt->second, bndVal);
	  ++fixed;
	}
      } else if (strcmp(bndType, "FR") == 0) {
	/* free variable */
	if (!checkBoundRow(line, tokens, 2, bndVar, varMap, varIt))
	  return -1;
	else {
	  lp.setxlFree(varIt->second);
	  lp.setxuFree(varIt->second);
	  ++lp.cfree;
	}
      } else if (strcmp(bndType, "MI") == 0) {
	/* variable unbounded from below */
	/* Specifying this and no upper bound results in a free variable too!!!
	 * Using this would invalidate at least the debug vector <permFree>...
	 * */
	if (!checkBoundRow(line, tokens, 2, bndVar, varMap, varIt))
	  return -1;
	else
	  lp.setxlFree(varIt->second);
      } else if (strcmp(bndType, "PL") == 0) {
	/* positive variable */
	if (!checkBoundRow(line, tokens, 2, bndVar, varMap, varIt))
	  return -1;
      } else {
	/* unknown bound type */
	PRINTF("WARNING! ignoring unknown bound type %s in line:\n"
	    "%s\n", bndType, line);
      }
      if (!getLine(line, in, fSilent)) return -1;
    }
    if (!fSilent) PRINTF("BOUNDS\tread: %d fixed, %d free\n", fixed, lp.cfree);
  }
#ifdef DEBUG_OUTPUT
  lp.printBounds();
#endif

  /* 6) read ENDATA 
   * */
  if (std::strncmp(line, "ENDATA", 6) != 0) {
    /* ENDATA section should start with "ENDATA" in column 1 */
    PRINTF("ERROR! ENDATA expected in line:\n"
	      "%s\n", line);
    return -1;
  }
  if (!fSilent) PRINTF("ENDATA\tread\n");

  /* cleanup */
  for (tConMap::iterator it = conMap.begin();
       it != conMap.end();
       ++it)
    delete [] it->first;
  conMap.clear();
  for (tVarMap::iterator it = varMap.begin();
       it != varMap.end();
       ++it)
    delete [] it->first;
  varMap.clear();

  return 0;
}

int
readMps(char *sz, CLp &lp, bool fSilent = false)
{
  FILE *in;
  
  if (NULL == (in = fopen(sz, "r"))) {
    PRINTF(strerror(errno));
    PRINTF("\nreadMps: Cannot open MPS file %s\n", sz);
    return -1;
  }

  return readMps(in, lp, fSilent);
}

void
buildABc(CLpDouble &lp)
{
  /* write data to model structure
   *  a) A and B matrices (sparse)
   *  b) c vector (sparse)
   * */
  lp.A.nz = lp.ineMap.size();
  /* at least allocate 1 unit of memory - otherwise we get segmentation faults
   * if the number of inequality constraints is 0 */
  lp.A.pr = (double *) CALLOC((lp.A.nz >= 1 ? lp.A.nz : 1), sizeof(double));
  lp.A.ir = (int *) CALLOC((lp.A.nz >= 1 ? lp.A.nz : 1), sizeof(int));
  lp.A.jc = (int *) CALLOC(lp.n + 1, sizeof(int));
  lp.A.jc[0] = 0;
  int col = 1;
  int elm = 0;
  double *prW = lp.A.pr;
  int *irW = lp.A.ir;
  int *jcW = lp.A.jc + 1;
  for (ELMMAP(double)::iterator it = lp.ineMap.begin();
       it != lp.ineMap.end();
       ++it, ++elm) {
    while (col != it->first.second) {
      /* there are no more elements in the current column */
      *(jcW++) = elm;
      ++col;
    }
    *(irW++) = it->first.first - 1;
    *(prW++) = (double) it->second;
  }
  while (col++ <= lp.n) {
    *(jcW++) = elm;
  }
#ifdef DEBUG_OUTPUT
  printArray(lp.A.pr, lp.A.nz);
#endif
  lp.B.nz = lp.equMap.size();
  /* at least allocate 1 unit of memory - otherwise we get segmentation faults
   * if the number of equality constraints is 0 */
  lp.B.pr = (double *) CALLOC((lp.B.nz >= 1 ? lp.B.nz : 1), sizeof(double));
  lp.B.ir = (int *) CALLOC((lp.B.nz >= 1 ? lp.B.nz : 1), sizeof(int));
  lp.B.jc = (int *) CALLOC(lp.n + 1, sizeof(int));
  lp.B.jc[0] = 0;
  col = 1;
  elm = 0;
  prW = lp.B.pr;
  irW = lp.B.ir;
  jcW = lp.B.jc + 1;
  for (ELMMAP(double)::iterator it = lp.equMap.begin();
       it != lp.equMap.end();
       ++it, ++elm) {
    while (col != it->first.second) {
      /* there are no more elements in the current column */
      *(jcW++) = elm;
      ++col;
    }
    *(irW++) = it->first.first - 1;
    *(prW++) = it->second;
  }
  while (col++ <= lp.n) {
    *(jcW++) = elm;
  }
  lp.c = (double *) CALLOC(lp.n, sizeof(double));
  for (SVECTOR(double)::iterator it = lp.vObjective.begin();
       it != lp.vObjective.end();
       ++it)
    lp.c[it->first - 1] = it->second;
}

#endif
